@@ -12,7 +12,6 @@
 #include <linux/time.h>
 #include <linux/kernel.h>
 #include <linux/mm.h>
-#include <linux/vmalloc.h>
 #include <linux/string.h>
 #include <linux/stat.h>
 #include <linux/errno.h>
@@ -39,7 +38,7 @@
 #include <linux/sunrpc/bc_xprt.h>
 #include <linux/nsproxy.h>
 #include <linux/pid_namespace.h>
-
+#include <linux/vmalloc.h>
 
 #include "nfs4_fs.h"
 #include "callback.h"
@@ -75,13 +74,6 @@ static DEFINE_SPINLOCK(nfs_version_lock);
 static DEFINE_MUTEX(nfs_version_mutex);
 static LIST_HEAD(nfs_versions);
 static DEFINE_IDA(nfs_client_trids);
-static DEFINE_IDA(nfs_server_trids);
-
-void nfs_client_cleanup_trids(void)
-{
-	ida_destroy(&nfs_client_trids);
-	ida_destroy(&nfs_server_trids);
-}
 
 static int nfs_client_alloc_trid(struct nfs_client *clnt)
 {
@@ -98,21 +90,6 @@ static void nfs_client_free_trid(struct nfs_client *clnt)
 	ida_simple_remove(&nfs_client_trids, clnt->cl_trid);
 }
 
-static int nfs_server_alloc_trid(struct nfs_server *server)
-{
-	int id = ida_simple_get(&nfs_server_trids, 4000, 0, GFP_KERNEL);
-	if (id < 0)
-		return id;
-
-	server->trid = id;
-	return 0;
-}
-
-static void nfs_server_free_trid(struct nfs_server *server)
-{
-	ida_simple_remove(&nfs_server_trids, server->trid);
-}
-
 /*
  * RPC cruft for NFS
  */
@@ -127,7 +104,6 @@ const struct rpc_program nfs_program = {
 	.number			= NFS_PROGRAM,
 	.nrvers			= ARRAY_SIZE(nfs_version),
 	.version		= nfs_version,
-	.stats			= &nfs_rpcstat,
 	.pipe_dir_name		= NFS_PIPE_DIRNAME,
 };
 
@@ -258,13 +234,11 @@ struct nfs_client *nfs_alloc_client(const struct nfs_client_initdata *cl_init)
 	clp->cl_proto = cl_init->proto;
 	clp->cl_nconnect = cl_init->nconnect;
 	clp->cl_max_connect = cl_init->max_connect ? cl_init->max_connect : 1;
-	clp->cl_pconnect = cl_init->pconnect ? cl_init->pconnect : 1;
 	clp->cl_net = get_net(cl_init->net);
 	clp->cl_xprt_id = cl_init->xprt_id;
 
 	clp->cl_principal = "*";
-	nfs_fscache_get_client_cookie(clp);
-
+	clp->cl_xprtsec = cl_init->xprtsec;
 	nfs_client_debugfs_register(clp);
 	trace_nfs_client_new(clp);
 
@@ -329,7 +303,6 @@ void nfs_free_client(struct nfs_client *clp)
 {
 	trace_nfs_client_free_start(clp);
 	nfs_client_debugfs_unregister(clp);
-	nfs_fscache_release_client_cookie(clp);
 
 	/* -EIO all pending I/O */
 	if (!IS_ERR(clp->cl_rpcclient))
@@ -344,6 +317,7 @@ void nfs_free_client(struct nfs_client *clp)
 	kfree(clp->cl_hostname);
 	kfree(clp->cl_acceptor);
 	trace_nfs_client_free_done(clp);
+
 	nfs_client_free_trid(clp);
 	kfree(clp);
 }
@@ -380,7 +354,7 @@ EXPORT_SYMBOL_GPL(nfs_put_client);
 static struct nfs_client *nfs_match_client(const struct nfs_client_initdata *data)
 {
 	struct nfs_client *clp;
-	const struct sockaddr *sap = data->addr;
+	const struct sockaddr *sap = (struct sockaddr *)data->addr;
 	struct nfs_net *nn = net_generic(data->net, nfs_net_id);
 	int error;
 
@@ -441,6 +415,10 @@ again:
 							   sap))
 				continue;
 
+		/* Match the xprt security policy */
+		if (clp->cl_xprtsec.policy != data->xprtsec.policy)
+			continue;
+
 		refcount_inc(&clp->cl_count);
 		return clp;
 	}
@@ -573,6 +551,7 @@ void nfs_init_timeout_values(struct rpc_timeout *to, int proto,
 
 	switch (proto) {
 	case XPRT_TRANSPORT_TCP:
+	case XPRT_TRANSPORT_TCP_TLS:
 	case XPRT_TRANSPORT_RDMA:
 		if (retrans == NFS_UNSPEC_RETRANS)
 			to->to_retries = NFS_DEF_TCP_RETRANS;
@@ -611,6 +590,7 @@ int nfs_create_rpc_client(struct nfs_client *clp,
 			  const struct nfs_client_initdata *cl_init,
 			  rpc_authflavor_t flavor)
 {
+	struct nfs_net		*nn = net_generic(clp->cl_net, nfs_net_id);
 	struct rpc_clnt		*clnt = NULL;
 	struct rpc_create_args args = {
 		.net		= clp->cl_net,
@@ -631,16 +611,17 @@ int nfs_create_rpc_client(struct nfs_client *clp,
 		.servername	= clp->cl_hostname,
 		.nodename	= cl_init->nodename,
 		.program	= &nfs_program,
+		.stats		= &nn->rpcstats,
 		.version	= clp->rpc_ops->version,
 		.authflavor	= flavor,
 		.cred		= cl_init->cred,
+		.xprtsec	= cl_init->xprtsec,
+		.connect_timeout = cl_init->connect_timeout,
+		.reconnect_timeout = cl_init->reconnect_timeout,
 		.flags		= clp->cl_no_idle_xprt ?
 			RPC_CLNT_CREATE_NO_IDLE_TIMEOUT : 0,
 	};
 
-	if (args.version == 3)
-		args.pconnect = clp->cl_pconnect;
-
 	if (test_bit(NFS_CS_DISCRTRY, &clp->cl_flags))
 		args.flags |= RPC_CLNT_CREATE_DISCRTRY;
 	if (test_bit(NFS_CS_NO_RETRANS_TIMEOUT, &clp->cl_flags))
@@ -721,6 +702,7 @@ static int nfs_start_lockd(struct nfs_server *server)
 
 	server->nlm_host = host;
 	server->destroy = nfs_destroy_server;
+	nfs_sysfs_link_rpc_client(server, nlmclnt_rpc_clnt(host), NULL);
 	return 0;
 }
 
@@ -752,6 +734,7 @@ int nfs_init_server_rpcclient(struct nfs_server *server,
 	if (server->flags & NFS_MOUNT_SOFT)
 		server->client->cl_softrtry = 1;
 
+	nfs_sysfs_link_rpc_client(server, server->client, NULL);
 	return 0;
 }
 EXPORT_SYMBOL_GPL(nfs_init_server_rpcclient);
@@ -797,7 +780,7 @@ static int nfs_init_server(struct nfs_server *server,
 	struct rpc_timeout timeparms;
 	struct nfs_client_initdata cl_init = {
 		.hostname = ctx->nfs_server.hostname,
-		.addr = (const struct sockaddr *)&ctx->nfs_server.address,
+		.addr = &ctx->nfs_server._address,
 		.addrlen = ctx->nfs_server.addrlen,
 		.nfs_mod = ctx->nfs_mod,
 		.proto = ctx->nfs_server.protocol,
@@ -805,8 +788,8 @@ static int nfs_init_server(struct nfs_server *server,
 		.timeparms = &timeparms,
 		.cred = server->cred,
 		.nconnect = ctx->nfs_server.nconnect,
-		.pconnect = ctx->nfs_server.pconnect,
 		.init_flags = (1UL << NFS_CS_REUSEPORT),
+		.xprtsec = ctx->xprtsec,
 		.local_ports = ctx->local_ports,
 		.local_ports_failover = ctx->local_ports_failover,
 		.remote_ports = ctx->remote_ports,
@@ -814,6 +797,7 @@ static int nfs_init_server(struct nfs_server *server,
 		.remote_ports_offset_provided = ctx->nfs_server.remoteports_offset_provided,
 		.spread_reads = ctx->nfs_server.spread_reads,
 		.spread_writes = ctx->nfs_server.spread_writes,
+		.no_idle_xprt = ctx->no_idle_xprt,
 		.mdconnect = ctx->nfs_server.mdconnect,
 	};
 	struct nfs_client *clp;
@@ -826,23 +810,20 @@ static int nfs_init_server(struct nfs_server *server,
 	if (ctx->flags & NFS_MOUNT_NOSHARE_XPRT)
 		set_bit(NFS_CS_NO_SHARE, &cl_init.init_flags);
 
+	/* Allocate or find a client reference we can use */
+	clp = nfs_get_client(&cl_init);
+	if (IS_ERR(clp))
+		return PTR_ERR(clp);
+
 	if (cl_init.mdconnect + cl_init.nconnect > NFS_MAX_CONNECTIONS) {
 		cl_init.mdconnect = NFS_MAX_CONNECTIONS - cl_init.nconnect;
 		printk(KERN_WARNING "NFS: Clamping down mdconnect to '%d'\n",
 		       cl_init.mdconnect);
 	}
 
-	cl_init.no_idle_xprt = get_no_idle_xprt(ctx);
-
-	if (ctx->local_ports && ctx->local_ports->nr > ctx->nfs_server.nconnect)
-		printk(KERN_WARNING "NFS: nconnect value is lower than available local ports\n");
-
-	/* Allocate or find a client reference we can use */
-	clp = nfs_get_client(&cl_init);
-	if (IS_ERR(clp))
-		return PTR_ERR(clp);
-
 	server->nfs_client = clp;
+	nfs_sysfs_add_server(server);
+	nfs_sysfs_link_rpc_client(server, clp->cl_rpcclient, "_state");
 
 	/* Initialise the client representation from the mount data */
 	server->flags = ctx->flags;
@@ -861,9 +842,9 @@ static int nfs_init_server(struct nfs_server *server,
 	}
 
 	if (ctx->rsize)
-		server->rsize = nfs_block_size(ctx->rsize, NULL);
+		server->rsize = nfs_io_size(ctx->rsize, clp->cl_proto);
 	if (ctx->wsize)
-		server->wsize = nfs_block_size(ctx->wsize, NULL);
+		server->wsize = nfs_io_size(ctx->wsize, clp->cl_proto);
 
 	server->acregmin = ctx->acregmin * HZ;
 	server->acregmax = ctx->acregmax * HZ;
@@ -910,18 +891,19 @@ error:
 static void nfs_server_set_fsinfo(struct nfs_server *server,
 				  struct nfs_fsinfo *fsinfo)
 {
+	struct nfs_client *clp = server->nfs_client;
 	unsigned long max_rpc_payload, raw_max_rpc_payload;
 
 	/* Work out a lot of parameters */
 	if (server->rsize == 0)
-		server->rsize = nfs_block_size(fsinfo->rtpref, NULL);
+		server->rsize = nfs_io_size(fsinfo->rtpref, clp->cl_proto);
 	if (server->wsize == 0)
-		server->wsize = nfs_block_size(fsinfo->wtpref, NULL);
+		server->wsize = nfs_io_size(fsinfo->wtpref, clp->cl_proto);
 
 	if (fsinfo->rtmax >= 512 && server->rsize > fsinfo->rtmax)
-		server->rsize = nfs_block_size(fsinfo->rtmax, NULL);
+		server->rsize = nfs_io_size(fsinfo->rtmax, clp->cl_proto);
 	if (fsinfo->wtmax >= 512 && server->wsize > fsinfo->wtmax)
-		server->wsize = nfs_block_size(fsinfo->wtmax, NULL);
+		server->wsize = nfs_io_size(fsinfo->wtmax, clp->cl_proto);
 
 	raw_max_rpc_payload = rpc_max_payload(server->client);
 	max_rpc_payload = nfs_block_size(raw_max_rpc_payload, NULL);
@@ -979,7 +961,7 @@ static void nfs_server_set_fsinfo(struct nfs_server *server,
 /*
  * Probe filesystem information, including the FSID on v2/v3
  */
-int nfs_probe_fsinfo(struct nfs_server *server, struct nfs_fh *mntfh, struct nfs_fattr *fattr)
+static int nfs_probe_fsinfo(struct nfs_server *server, struct nfs_fh *mntfh, struct nfs_fattr *fattr)
 {
 	struct nfs_fsinfo fsinfo;
 	struct nfs_client *clp = server->nfs_client;
@@ -1012,7 +994,8 @@ int nfs_probe_fsinfo(struct nfs_server *server, struct nfs_fh *mntfh, struct nfs
 	}
 
 	if (clp->rpc_ops->discover_trunking != NULL &&
-			(server->caps & NFS_CAP_FS_LOCATIONS)) {
+			(server->caps & NFS_CAP_FS_LOCATIONS &&
+			 (server->flags & NFS_MOUNT_TRUNK_DISCOVERY))) {
 		error = clp->rpc_ops->discover_trunking(server, mntfh);
 		if (error < 0)
 			return error;
@@ -1020,7 +1003,30 @@ int nfs_probe_fsinfo(struct nfs_server *server, struct nfs_fh *mntfh, struct nfs
 
 	return 0;
 }
-EXPORT_SYMBOL_GPL(nfs_probe_fsinfo);
+
+/*
+ * Grab the destination's particulars, including lease expiry time.
+ *
+ * Returns zero if probe succeeded and retrieved FSID matches the FSID
+ * we have cached.
+ */
+int nfs_probe_server(struct nfs_server *server, struct nfs_fh *mntfh)
+{
+	struct nfs_fattr *fattr;
+	int error;
+
+	fattr = nfs_alloc_fattr();
+	if (fattr == NULL)
+		return -ENOMEM;
+
+	/* Sanity: the probe won't work if the destination server
+	 * does not recognize the migrated FH. */
+	error = nfs_probe_fsinfo(server, mntfh, fattr);
+
+	nfs_free_fattr(fattr);
+	return error;
+}
+EXPORT_SYMBOL_GPL(nfs_probe_server);
 
 /*
  * Copy useful information when duplicating a server record
@@ -1074,17 +1080,23 @@ void nfs_server_remove_lists(struct nfs_server *server)
 }
 EXPORT_SYMBOL_GPL(nfs_server_remove_lists);
 
+static DEFINE_IDA(s_sysfs_ids);
+
 /*
  * Allocate and initialise a server record
  */
 struct nfs_server *nfs_alloc_server(void)
 {
 	struct nfs_server *server;
-	int err;
 
 	server = kzalloc(sizeof(struct nfs_server), GFP_KERNEL);
 	if (!server)
-		return NULL;
+		goto err_out;
+
+	server->s_sysfs_id = ida_alloc(&s_sysfs_ids, GFP_KERNEL);
+	if (server->s_sysfs_id < 0)
+		goto err_free;
+
 
 	server->client = server->client_acl = ERR_PTR(-EINVAL);
 
@@ -1095,21 +1107,16 @@ struct nfs_server *nfs_alloc_server(void)
 	INIT_LIST_HEAD(&server->layouts);
 	INIT_LIST_HEAD(&server->state_owners_lru);
 	INIT_LIST_HEAD(&server->ss_copies);
+	INIT_LIST_HEAD(&server->ss_src_copies);
 
 	atomic_set(&server->active, 0);
 
 	server->io_stats = nfs_alloc_iostats();
 	if (!server->io_stats)
-		goto err;
+		goto err_free_id;
 
 	server->change_attr_type = NFS4_CHANGE_TYPE_IS_UNDEFINED;
 
-	err = nfs_server_alloc_trid(server);
-	if (err < 0) {
-		nfs_free_iostats(server->io_stats);
-		goto err;
-	}
-
 	ida_init(&server->openowner_id);
 	ida_init(&server->lockowner_id);
 	pnfs_init_server(server);
@@ -1117,10 +1124,14 @@ struct nfs_server *nfs_alloc_server(void)
 
 	nfs_server_debugfs_register(server);
 	trace_nfs_server_new(server);
+
 	return server;
 
-err:
+err_free_id:
+	ida_free(&s_sysfs_ids, server->s_sysfs_id);
+err_free:
 	kfree(server);
+err_out:
 	return NULL;
 }
 EXPORT_SYMBOL_GPL(nfs_alloc_server);
@@ -1136,8 +1147,7 @@ void nfs_free_server(struct nfs_server *server)
 	nfs_server_debugfs_unregister(server);
 
 	for (i = 0; i < server->superblock.nr_alt_servers; i++) {
-		struct nfs_server *alt_server =
-			server->superblock.alt_server[i];
+		struct nfs_server *alt_server = server->superblock.alt_server[i];
 		nfs_free_server(alt_server);
 	}
 
@@ -1153,14 +1163,20 @@ void nfs_free_server(struct nfs_server *server)
 
 	nfs_put_client(server->nfs_client);
 
+	if (server->kobj.state_initialized) {
+		nfs_sysfs_remove_server(server);
+		kobject_put(&server->kobj);
+	}
 	ida_destroy(&server->lockowner_id);
 	ida_destroy(&server->openowner_id);
 	nfs_free_iostats(server->io_stats);
 	put_cred(server->cred);
-	trace_nfs_server_free_done(server);
-	nfs_server_free_trid(server);
 	if (server->remote_ports)
 		vfree(server->remote_ports);
+
+	trace_nfs_server_free_done(server);
+	ida_free(&s_sysfs_ids, server->s_sysfs_id);
+
 	kfree(server);
 	nfs_release_automount_timer();
 }
@@ -1209,7 +1225,7 @@ struct nfs_server *nfs_create_server(struct fs_context *fc)
 
 	if (!(fattr->valid & NFS_ATTR_FATTR)) {
 		error = ctx->nfs_mod->rpc_ops->getattr(server, ctx->mntfh,
-						       fattr, NULL, NULL);
+						       fattr, NULL);
 		if (error < 0) {
 			dprintk("nfs_create_server: getattr error = %d\n", -error);
 			goto error;
@@ -1252,8 +1268,8 @@ struct nfs_server *nfs_clone_server(struct nfs_server *source,
 				    struct nfs_fattr *fattr,
 				    rpc_authflavor_t flavor)
 {
-	struct nfs_server *server, *alt_server, *cloned;
-	struct nfs_fattr *fattr_fsinfo;
+	struct nfs_server *alt_server, *cloned;
+	struct nfs_server *server;
 	int error, i;
 
 	server = nfs_alloc_server();
@@ -1262,11 +1278,6 @@ struct nfs_server *nfs_clone_server(struct nfs_server *source,
 
 	server->cred = get_cred(source->cred);
 
-	error = -ENOMEM;
-	fattr_fsinfo = nfs_alloc_fattr();
-	if (fattr_fsinfo == NULL)
-		goto out_free_server;
-
 	/* Copy data from the source */
 	server->nfs_client = source->nfs_client;
 	server->destroy = source->destroy;
@@ -1275,6 +1286,11 @@ struct nfs_server *nfs_clone_server(struct nfs_server *source,
 
 	server->fsid = fattr->fsid;
 
+	nfs_sysfs_add_server(server);
+
+	nfs_sysfs_link_rpc_client(server,
+		server->nfs_client->cl_rpcclient, "_state");
+
 	error = nfs_init_server_rpcclient(server,
 			source->client->cl_timeout,
 			flavor);
@@ -1282,7 +1298,7 @@ struct nfs_server *nfs_clone_server(struct nfs_server *source,
 		goto out_free_server;
 
 	/* probe the filesystem info for this server filesystem */
-	error = nfs_probe_fsinfo(server, fh, fattr_fsinfo);
+	error = nfs_probe_server(server, fh);
 	if (error < 0)
 		goto out_free_server;
 
@@ -1298,23 +1314,17 @@ struct nfs_server *nfs_clone_server(struct nfs_server *source,
 	nfs_server_insert_lists(server);
 	server->mount_time = jiffies;
 
-	nfs_free_fattr(fattr_fsinfo);
-
 	for (i = 0; i < source->superblock.nr_alt_servers; i++) {
 		alt_server = source->superblock.alt_server[i];
 		cloned = nfs_clone_server(alt_server, fh, fattr, flavor);
-		if (IS_ERR(cloned))
-			continue;
-
-		nfs_server_add_alt(server, cloned);
+		if (IS_ERR_OR_NULL(cloned))
+			pr_warn("VastNFS cannot alocate NFSv4.x alt server\n");
+		else
+			nfs_server_add_alt(server, cloned);
 	}
-	server->remote_ports = source->remote_ports;
-	source->remote_ports = NULL;
-
 	return server;
 
 out_free_server:
-	nfs_free_fattr(fattr_fsinfo);
 	nfs_free_server(server);
 	return ERR_PTR(error);
 }
@@ -1331,6 +1341,8 @@ void nfs_clients_init(struct net *net)
 #endif
 	spin_lock_init(&nn->nfs_client_lock);
 	nn->boot_time = ktime_get_real();
+	memset(&nn->rpcstats, 0, sizeof(nn->rpcstats));
+	nn->rpcstats.program = &nfs_program;
 
 	nfs_netns_sysfs_setup(nn, net);
 }
@@ -1512,18 +1524,6 @@ static int nfs_volume_list_show(struct seq_file *m, void *v)
 	return 0;
 }
 
-PROC_CREATE_NET_COMPAT(nfs_proc_create_net_servers,
-		       nfs_server_list_fops,
-		       nfs_server_list_ops,
-		       nfs_server_list_open,
-		       sizeof(struct seq_net_private));
-
-PROC_CREATE_NET_COMPAT(nfs_proc_create_net_volumes,
-		       nfs_volume_list_fops,
-		       nfs_volume_list_ops,
-		       nfs_volume_list_open,
-		       sizeof(struct seq_net_private));
-
 int nfs_fs_proc_net_init(struct net *net)
 {
 	struct nfs_net *nn = net_generic(net, nfs_net_id);
@@ -1534,12 +1534,14 @@ int nfs_fs_proc_net_init(struct net *net)
 		goto error_0;
 
 	/* a file of servers with which we're dealing */
-	p = nfs_proc_create_net_servers("servers", S_IFREG|S_IRUGO, nn->proc_nfsfs);
+	p = proc_create_net("servers", S_IFREG|S_IRUGO, nn->proc_nfsfs,
+			&nfs_server_list_ops, sizeof(struct seq_net_private));
 	if (!p)
 		goto error_1;
 
 	/* a file of volumes that we have mounted */
-	p = nfs_proc_create_net_volumes("volumes", S_IFREG|S_IRUGO, nn->proc_nfsfs);
+	p = proc_create_net("volumes", S_IFREG|S_IRUGO, nn->proc_nfsfs,
+			&nfs_volume_list_ops, sizeof(struct seq_net_private));
 	if (!p)
 		goto error_1;
 	return 0;
@@ -1584,6 +1586,8 @@ error_0:
 void nfs_fs_proc_exit(void)
 {
 	remove_proc_subtree("fs/nfsfs", NULL);
+	ida_destroy(&nfs_client_trids);
+	ida_destroy(&s_sysfs_ids);
 }
 
 #endif /* CONFIG_PROC_FS */
