@@ -17,10 +17,9 @@
 #define NFS_MAX_CONNECTIONS 64
 #define NFS_MAX_MD_CONNECTIONS  8
 
-#define NFS_SB_MASK (SB_RDONLY|SB_NOSUID|SB_NODEV|SB_NOEXEC|SB_SYNCHRONOUS)
+#define NFS_SB_MASK (SB_NOSUID|SB_NODEV|SB_NOEXEC|SB_SYNCHRONOUS)
 
 extern const struct export_operations nfs_export_ops;
-extern struct workqueue_struct *nfs_async_write_workqueue;
 
 struct nfs_string;
 struct nfs_pageio_descriptor;
@@ -76,7 +75,7 @@ static inline fmode_t flags_to_mode(int flags)
 struct nfs_client_initdata {
 	unsigned long init_flags;
 	const char *hostname;			/* Hostname of the server */
-	const struct sockaddr *addr;		/* Address of the server */
+	const struct sockaddr_storage *addr;	/* Address of the server */
 	const char *nodename;			/* Hostname of the client */
 	const char *ip_addr;			/* IP address of the client */
 	size_t addrlen;
@@ -94,11 +93,13 @@ struct nfs_client_initdata {
 	u32 minorversion;
 	unsigned int nconnect;
 	unsigned int max_connect;
-	unsigned int pconnect;
-	u32 server_idx;				/* NFSv4.1 server multipath index */
+	u32 server_idx;				/* NFSv4.x server multipath index */
 	struct net *net;
 	const struct rpc_timeout *timeparms;
 	const struct cred *cred;
+	struct xprtsec_parms xprtsec;
+	unsigned long connect_timeout;
+	unsigned long reconnect_timeout;
 	unsigned int xprt_id;
 };
 
@@ -120,6 +121,7 @@ struct nfs_fs_context {
 	unsigned int		bsize;
 	struct nfs_auth_info	auth_info;
 	rpc_authflavor_t	selected_flavor;
+	struct xprtsec_parms	xprtsec;
 	char			*client_address;
 	unsigned int		version;
 	unsigned int		minorversion;
@@ -127,10 +129,10 @@ struct nfs_fs_context {
 	char			*fscache_uniq;
 	unsigned short		protofamily;
 	unsigned short		mountfamily;
-	bool			has_sec_mnt_opts;
+	bool                    has_sec_mnt_opts;
 	struct rpc_portgroup	*local_ports;
 	struct rpc_portgroup	*remote_ports;
-	int			no_idle_xprt;
+	bool			no_idle_xprt;
 	bool			local_ports_failover;
 	u32			server_idx;
 	unsigned int		xprt_id;
@@ -160,7 +162,6 @@ struct nfs_fs_context {
 		unsigned short		nconnect;
 		unsigned short		mdconnect;
 		unsigned short		max_connect;
-		unsigned short		pconnect;
 		unsigned short		export_path_len;
 		u32			remoteports_offset;
 		bool			remoteports_offset_provided;
@@ -185,40 +186,34 @@ extern void
 nfs_fs_context_trace(const char *kind, struct fs_context *fc,
 		     const char *fmt, ...);
 
-#define nfs_errorf(fc, fmt, ...) \
-	({nfs_fs_context_trace("error", fc, fmt, ## __VA_ARGS__); \
-	(compat_fs_context_to_log(fc) ?						\
-	errorf(fc, fmt, ## __VA_ARGS__) :			\
+#define nfs_errorf(fc, fmt, ...)					\
+	({nfs_fs_context_trace("error", fc, fmt, ## __VA_ARGS__);	\
+	((fc)->log.log ? errorf(fc, fmt, ## __VA_ARGS__) :		\
 	({ dprintk(fmt "\n", ## __VA_ARGS__); })); })
 
-#define nfs_ferrorf(fc, fac, fmt, ...) \
-	({nfs_fs_context_trace("ferror", fc, fmt, ## __VA_ARGS__); \
-	(compat_fs_context_to_log(fc) ?		\
-	errorf(fc, fmt, ## __VA_ARGS__) :			\
+#define nfs_ferrorf(fc, fac, fmt, ...)					\
+	({nfs_fs_context_trace("ferror", fc, fmt, ## __VA_ARGS__);	\
+	((fc)->log.log ? errorf(fc, fmt, ## __VA_ARGS__) :		\
 	({ dfprintk(fac, fmt "\n", ## __VA_ARGS__); })); })
 
-#define nfs_invalf(fc, fmt, ...) \
-	({nfs_fs_context_trace("invalf", fc, fmt, ## __VA_ARGS__); \
-	(compat_fs_context_to_log(fc) ?		\
-	invalf(fc, fmt, ## __VA_ARGS__) :			\
-	({ dprintk(fmt "\n", ## __VA_ARGS__);  -EINVAL; })); })
+#define nfs_invalf(fc, fmt, ...)					\
+	({nfs_fs_context_trace("invalf", fc, fmt, ## __VA_ARGS__);	\
+	((fc)->log.log ? invalf(fc, fmt, ## __VA_ARGS__) :		\
+	({ dprintk(fmt "\n", ## __VA_ARGS__); -EINVAL; })); })
 
-#define nfs_finvalf(fc, fac, fmt, ...) \
-	({nfs_fs_context_trace("finvalf", fc, fmt, ## __VA_ARGS__); \
-	(compat_fs_context_to_log(fc) ?		\
-	invalf(fc, fmt, ## __VA_ARGS__) :			\
-	({ dfprintk(fac, fmt "\n", ## __VA_ARGS__);  -EINVAL; })); })
+#define nfs_finvalf(fc, fac, fmt, ...)					\
+	({nfs_fs_context_trace("finvalf", fc, fmt, ## __VA_ARGS__);	\
+	((fc)->log.log ? invalf(fc, fmt, ## __VA_ARGS__) :		\
+	({ dfprintk(fac, fmt "\n", ## __VA_ARGS__); -EINVAL; })); })
 
-#define nfs_warnf(fc, fmt, ...) \
-	({nfs_fs_context_trace("warnf", fc, fmt, ##__VA_ARGS__); \
-	(compat_fs_context_to_log(fc) ?		\
-	warnf(fc, fmt, ## __VA_ARGS__) :			\
+#define nfs_warnf(fc, fmt, ...)						\
+	({nfs_fs_context_trace("warnf", fc, fmt, ##__VA_ARGS__);	\
+	((fc)->log.log ? warnf(fc, fmt, ## __VA_ARGS__) :		\
 	({ dprintk(fmt "\n", ## __VA_ARGS__); })); })
 
-#define nfs_fwarnf(fc, fac, fmt, ...) \
-	({nfs_fs_context_trace("warn", fc, fmt, ## __VA_ARGS__); \
-	(compat_fs_context_to_log(fc) ?		\
-	warnf(fc, fmt, ## __VA_ARGS__) :			\
+#define nfs_fwarnf(fc, fac, fmt, ...)					\
+	({nfs_fs_context_trace("warn", fc, fmt, ## __VA_ARGS__)		\
+	((fc)->log.log ? warnf(fc, fmt, ## __VA_ARGS__) :		\
 	({ dfprintk(fac, fmt "\n", ## __VA_ARGS__); })); })
 
 static inline struct nfs_fs_context *nfs_fc2context(const struct fs_context *fc)
@@ -228,7 +223,7 @@ static inline struct nfs_fs_context *nfs_fc2context(const struct fs_context *fc)
 
 /* mount_clnt.c */
 struct nfs_mount_request {
-	struct sockaddr		*sap;
+	struct sockaddr_storage	*sap;
 	size_t			salen;
 	char			*hostname;
 	char			*dirpath;
@@ -251,7 +246,7 @@ extern void nfs_clients_exit(struct net *net);
 extern struct nfs_client *nfs_alloc_client(const struct nfs_client_initdata *);
 int nfs_create_rpc_client(struct nfs_client *, const struct nfs_client_initdata *, rpc_authflavor_t);
 struct nfs_client *nfs_get_client(const struct nfs_client_initdata *);
-int nfs_probe_fsinfo(struct nfs_server *server, struct nfs_fh *, struct nfs_fattr *);
+int nfs_probe_server(struct nfs_server *, struct nfs_fh *);
 void nfs_server_insert_lists(struct nfs_server *);
 void nfs_server_remove_lists(struct nfs_server *);
 void nfs_init_timeout_values(struct rpc_timeout *to, int proto, int timeo, int retrans);
@@ -267,12 +262,13 @@ extern struct nfs_client *
 nfs4_find_client_sessionid(struct net *, const struct sockaddr *,
 				struct nfs4_sessionid *, u32);
 extern struct nfs_server *nfs_create_server(struct fs_context *);
+extern void nfs4_server_set_init_caps(struct nfs_server *);
 extern struct nfs_server *nfs4_create_server(struct fs_context *);
 extern struct nfs_server *nfs4_create_multi_server(struct fs_context *);
 extern struct nfs_server *nfs4_create_referral_server(struct fs_context *);
 extern void nfs_server_add_alt(struct nfs_server *parent, struct nfs_server *srv);
 extern int nfs4_update_server(struct nfs_server *server, const char *hostname,
-					struct sockaddr *sap, size_t salen,
+					struct sockaddr_storage *sap, size_t salen,
 					struct net *net);
 extern void nfs_free_server(struct nfs_server *server);
 extern struct nfs_server *nfs_clone_server(struct nfs_server *,
@@ -284,7 +280,7 @@ extern int nfs_client_init_status(const struct nfs_client *clp);
 extern int nfs_wait_client_init_complete(const struct nfs_client *clp);
 extern void nfs_mark_client_ready(struct nfs_client *clp, int state);
 extern struct nfs_client *nfs4_set_ds_client(struct nfs_server *mds_srv,
-					     const struct sockaddr *ds_addr,
+					     const struct sockaddr_storage *ds_addr,
 					     int ds_addrlen, int ds_proto,
 					     unsigned int ds_timeo,
 					     unsigned int ds_retrans,
@@ -292,7 +288,7 @@ extern struct nfs_client *nfs4_set_ds_client(struct nfs_server *mds_srv,
 extern struct rpc_clnt *nfs4_find_or_create_ds_client(struct nfs_client *,
 						struct inode *);
 extern struct nfs_client *nfs3_set_ds_client(struct nfs_server *mds_srv,
-			const struct sockaddr *ds_addr, int ds_addrlen,
+			const struct sockaddr_storage *ds_addr, int ds_addrlen,
 			int ds_proto, unsigned int ds_timeo,
 			unsigned int ds_retrans);
 #ifdef CONFIG_PROC_FS
@@ -323,7 +319,6 @@ extern const struct svc_version nfs4_callback_version4;
 
 /* fs_context.c */
 extern struct file_system_type nfs_fs_type;
-COMPAT_FS_CONTEXT_DETECTOR_DECL(nfs_fs_type);
 
 /* pagelist.c */
 extern int __init nfs_init_nfspagecache(void);
@@ -392,13 +387,32 @@ nfs4_label_copy(struct nfs4_label *dst, struct nfs4_label *src)
 	if (!dst || !src)
 		return NULL;
 
+#ifdef COMPAT_DETECT_SECURITY_UBUNTU_LSMCONTEXT
+	if (src->lsmctx.len > NFS4_MAXLABELLEN)
+		return NULL;
+#else /* !COMPAT_DETECT_SECURITY_UBUNTU_LSMCONTEXT */
 	if (src->len > NFS4_MAXLABELLEN)
 		return NULL;
+#endif /* COMPAT_DETECT_SECURITY_UBUNTU_LSMCONTEXT */
 
 	dst->lfs = src->lfs;
 	dst->pi = src->pi;
+
+#ifdef COMPAT_DETECT_SECURITY_UBUNTU_LSMCONTEXT
+	/* Use an invalid LSM slot as lsmctx should never be "released" */
+
+#ifdef COMPAT_DETECT_SECURITY_LSMCONTEXT_USES_ID
+	dst->lsmctx.id = -1;
+#else /* !COMPAT_DETECT_SECURITY_LSMCONTEXT_USES_ID */
+	dst->lsmctx.slot = -1;
+#endif /* COMPAT_DETECT_SECURITY_LSMCONTEXT_USES_ID */
+
+	dst->lsmctx.len = src->lsmctx.len;
+	memcpy(dst->lsmctx.context, src->lsmctx.context, src->lsmctx.len);
+#else /* !COMPAT_DETECT_SECURITY_UBUNTU_LSMCONTEXT */
 	dst->len = src->len;
 	memcpy(dst->label, src->label, src->len);
+#endif /* COMPAT_DETECT_SECURITY_UBUNTU_LSMCONTEXT */
 
 	return dst;
 }
@@ -426,37 +440,59 @@ extern struct nfs_client *nfs_init_client(struct nfs_client *clp,
 			   const struct nfs_client_initdata *);
 
 /* dir.c */
-extern void nfs_advise_use_readdirplus(struct inode *dir);
-extern void nfs_force_use_readdirplus(struct inode *dir);
+extern void nfs_readdir_record_entry_cache_hit(struct inode *dir);
+extern void nfs_readdir_record_entry_cache_miss(struct inode *dir);
 extern unsigned long nfs_access_cache_count(struct shrinker *shrink,
 					    struct shrink_control *sc);
 extern unsigned long nfs_access_cache_scan(struct shrinker *shrink,
 					   struct shrink_control *sc);
 struct dentry *nfs_lookup(struct inode *, struct dentry *, unsigned int);
-int nfs_create(COMPAT_INIT_USER_NS_PARAM struct inode *, struct dentry *,
+void nfs_d_prune_case_insensitive_aliases(struct inode *inode);
+int nfs_create(COMPAT_STRUCT_MNT_IDMAP *idmap, struct inode *, struct dentry *,
 	       umode_t, bool);
-int nfs_mkdir(COMPAT_INIT_USER_NS_PARAM struct inode *, struct dentry *,
+#ifdef COMPAT_DETECT_INODE_OP_MKDIR_RETURNS_DENTRY_6_15
+struct dentry *nfs_mkdir(COMPAT_STRUCT_MNT_IDMAP *idmap, struct inode *, struct dentry *,
+			 umode_t);
+#else
+int nfs_mkdir(COMPAT_STRUCT_MNT_IDMAP *idmap, struct inode *, struct dentry *,
 	      umode_t);
+#endif
 int nfs_rmdir(struct inode *, struct dentry *);
 int nfs_unlink(struct inode *, struct dentry *);
-int nfs_symlink(COMPAT_INIT_USER_NS_PARAM struct inode *, struct dentry *,
+int nfs_symlink(COMPAT_STRUCT_MNT_IDMAP *idmap, struct inode *, struct dentry *,
 		const char *);
 int nfs_link(struct dentry *, struct inode *, struct dentry *);
-int nfs_mknod(COMPAT_INIT_USER_NS_PARAM struct inode *, struct dentry *, umode_t,
+int nfs_mknod(COMPAT_STRUCT_MNT_IDMAP *idmap, struct inode *, struct dentry *, umode_t,
 	      dev_t);
-int nfs_rename(COMPAT_INIT_USER_NS_PARAM struct inode *, struct dentry *,
+int nfs_rename(COMPAT_STRUCT_MNT_IDMAP *idmap, struct inode *, struct dentry *,
 	       struct inode *, struct dentry *, unsigned int);
 
+#ifdef CONFIG_NFS_V4_2
+static inline __u32 nfs_access_xattr_mask(const struct nfs_server *server)
+{
+	if (!(server->caps & NFS_CAP_XATTR))
+		return 0;
+	return NFS4_ACCESS_XAREAD | NFS4_ACCESS_XAWRITE | NFS4_ACCESS_XALIST;
+}
+#else
+static inline __u32 nfs_access_xattr_mask(const struct nfs_server *server)
+{
+	return 0;
+}
+#endif
+
 /* file.c */
 int nfs_file_fsync(struct file *file, loff_t start, loff_t end, int datasync);
 loff_t nfs_file_llseek(struct file *, loff_t, int);
 ssize_t nfs_file_read(struct kiocb *, struct iov_iter *);
-#if defined(COMPAT_DETECT_GENERIC_FILE_SPLICE_READ_6_5)
-#define nfs_file_splice_read generic_file_splice_read
-#else
+#ifdef COMPAT_DETECT_FILEMAP_SPLICE_READ_6_3
 ssize_t nfs_file_splice_read(struct file *in, loff_t *ppos, struct pipe_inode_info *pipe,
 			     size_t len, unsigned int flags);
+#else
+#define nfs_file_splice_read generic_file_splice_read
 #endif
+ssize_t nfs_file_splice_write(struct pipe_inode_info *pipe, struct file *out,
+				loff_t *opos, size_t len , unsigned flags);
 int nfs_file_mmap(struct file *, struct vm_area_struct *);
 ssize_t nfs_file_write(struct kiocb *, struct iov_iter *);
 int nfs_file_release(struct inode *, struct file *);
@@ -468,7 +504,6 @@ int nfs_check_flags(int);
 extern struct workqueue_struct *nfsiod_workqueue;
 extern struct inode *nfs_alloc_inode(struct super_block *sb);
 extern void nfs_free_inode(struct inode *);
-extern void nfs_destroy_inode(struct inode *);
 extern int nfs_write_inode(struct inode *, struct writeback_control *);
 extern int nfs_drop_inode(struct inode *);
 extern void nfs_clear_inode(struct inode *);
@@ -477,7 +512,6 @@ extern void nfs_zap_acl_cache(struct inode *inode);
 extern void nfs_set_cache_invalid(struct inode *inode, unsigned long flags);
 extern bool nfs_check_cache_invalid(struct inode *, unsigned long);
 extern int nfs_wait_bit_killable(struct wait_bit_key *key, int mode);
-extern int nfs_wait_atomic_killable(atomic_t *p, unsigned int mode);
 
 /* super.c */
 extern const struct super_operations nfs_sops;
@@ -486,8 +520,6 @@ int nfs_try_get_tree(struct fs_context *);
 int nfs_get_tree_common(struct fs_context *);
 void nfs_kill_super(struct super_block *);
 
-extern struct rpc_stat nfs_rpcstat;
-
 extern int __init register_nfs_fs(void);
 extern void __exit unregister_nfs_fs(void);
 extern bool nfs_sb_active(struct super_block *sb);
@@ -495,6 +527,10 @@ extern void nfs_sb_deactive(struct super_block *sb);
 extern int nfs_client_for_each_server(struct nfs_client *clp,
 				      int (*fn)(struct nfs_server *, void *),
 				      void *data);
+#ifdef CONFIG_NFS_FSCACHE
+extern const struct netfs_request_ops nfs_netfs_ops;
+#endif
+
 /* io.c */
 extern void nfs_start_io_read(struct inode *inode);
 extern void nfs_end_io_read(struct inode *inode);
@@ -524,9 +560,15 @@ extern int nfs4_get_rootfh(struct nfs_server *server, struct nfs_fh *mntfh, bool
 
 struct nfs_pgio_completion_ops;
 /* read.c */
+extern const struct nfs_pgio_completion_ops nfs_async_read_completion_ops;
 extern void nfs_pageio_init_read(struct nfs_pageio_descriptor *pgio,
 			struct inode *inode, bool force_mds,
 			const struct nfs_pgio_completion_ops *compl_ops);
+extern bool nfs_read_alloc_scratch(struct nfs_pgio_header *hdr, size_t size);
+extern int nfs_read_add_folio(struct nfs_pageio_descriptor *pgio,
+			       struct nfs_open_context *ctx,
+			       struct folio *folio);
+extern void nfs_pageio_complete_read(struct nfs_pageio_descriptor *pgio);
 extern void nfs_read_prepare(struct rpc_task *task, void *calldata);
 extern void nfs_pageio_reset_read_mds(struct nfs_pageio_descriptor *pgio);
 
@@ -545,7 +587,6 @@ extern void nfs_pageio_init_write(struct nfs_pageio_descriptor *pgio,
 			const struct nfs_pgio_completion_ops *compl_ops);
 extern void nfs_pageio_reset_write_mds(struct nfs_pageio_descriptor *pgio);
 extern void nfs_commit_free(struct nfs_commit_data *p);
-extern void nfs_write_prepare(struct rpc_task *task, void *calldata);
 extern void nfs_commit_prepare(struct rpc_task *task, void *calldata);
 extern int nfs_initiate_commit(struct rpc_clnt *clnt,
 			       struct nfs_commit_data *data,
@@ -591,7 +632,6 @@ void nfs_pageio_stop_mirroring(struct nfs_pageio_descriptor *pgio);
 int nfs_filemap_write_and_wait_range(struct address_space *mapping,
 		loff_t lstart, loff_t lend);
 
-void nfs_client_cleanup_trids(void);
 void nfs_fs_context_cleanup_trids(void);
 
 #ifdef CONFIG_NFS_V4_1
@@ -622,21 +662,12 @@ void nfs_clear_pnfs_ds_commit_verifiers(struct pnfs_ds_commit_info *cinfo)
 }
 #endif
 
-void nfs_async_write_flush(struct nfs_inode *nfsi, int count);
-
 #ifdef CONFIG_MIGRATION
-#if !defined(COMPAT_DETECT_PAGE_FOLIO_MIGRATE_PAGE_6_0)
-extern int nfs_migrate_page(struct address_space *,
-		struct page *, struct page *, enum migrate_mode);
-#else
 int nfs_migrate_folio(struct address_space *, struct folio *dst,
 		struct folio *src, enum migrate_mode);
-#endif
 #else
-#if defined(COMPAT_DETECT_PAGE_FOLIO_MIGRATE_PAGE_6_0)
 #define nfs_migrate_folio NULL
 #endif
-#endif
 
 static inline int
 nfs_write_verifier_cmp(const struct nfs_write_verifier *v1,
@@ -660,6 +691,31 @@ static inline gfp_t nfs_io_gfp_mask(void)
 	return GFP_KERNEL;
 }
 
+/*
+ * Special version of should_remove_suid() that ignores capabilities.
+ */
+static inline int nfs_should_remove_suid(const struct inode *inode)
+{
+	umode_t mode = inode->i_mode;
+	int kill = 0;
+
+	/* suid always must be killed */
+	if (unlikely(mode & S_ISUID))
+		kill = ATTR_KILL_SUID;
+
+	/*
+	 * sgid without any exec bits is just a mandatory locking mark; leave
+	 * it alone.  If some exec bits are set, it's a real sgid; kill it.
+	 */
+	if (unlikely((mode & S_ISGID) && (mode & S_IXGRP)))
+		kill |= ATTR_KILL_SGID;
+
+	if (unlikely(kill && S_ISREG(mode)))
+		return kill;
+
+	return 0;
+}
+
 /* unlink.c */
 extern struct rpc_task *
 nfs_async_rename(struct inode *old_dir, struct inode *new_dir,
@@ -670,7 +726,7 @@ extern int nfs_sillyrename(struct inode *dir, struct dentry *dentry);
 /* direct.c */
 void nfs_init_cinfo_from_dreq(struct nfs_commit_info *cinfo,
 			      struct nfs_direct_req *dreq);
-extern ssize_t nfs_dreq_bytes_left(struct nfs_direct_req *dreq);
+extern ssize_t nfs_dreq_bytes_left(struct nfs_direct_req *dreq, loff_t offset);
 
 /* nfs4proc.c */
 extern struct nfs_client *nfs4_init_client(struct nfs_client *clp,
@@ -727,9 +783,9 @@ unsigned long nfs_block_bits(unsigned long bsize, unsigned char *nrbitsp)
 	if ((bsize & (bsize - 1)) || nrbitsp) {
 		unsigned char	nrbits;
 
-		for (nrbits = 31; nrbits && !(bsize & (1 << nrbits)); nrbits--)
+		for (nrbits = 31; nrbits && !(bsize & (1UL << nrbits)); nrbits--)
 			;
-		bsize = 1 << nrbits;
+		bsize = 1UL << nrbits;
 		if (nrbitsp)
 			*nrbitsp = nrbits;
 	}
@@ -760,6 +816,22 @@ unsigned long nfs_block_size(unsigned long bsize, unsigned char *nrbitsp)
 	return nfs_block_bits(bsize, nrbitsp);
 }
 
+/*
+ * Compute and set NFS server rsize / wsize
+ */
+static inline
+unsigned long nfs_io_size(unsigned long iosize, enum xprt_transports proto)
+{
+	if (iosize < NFS_MIN_FILE_IO_SIZE)
+		iosize = NFS_DEF_FILE_IO_SIZE;
+	else if (iosize >= NFS_MAX_FILE_IO_SIZE)
+		iosize = NFS_MAX_FILE_IO_SIZE;
+
+	if (proto == XPRT_TRANSPORT_UDP || iosize < PAGE_SIZE)
+		return nfs_block_bits(iosize, NULL);
+	return iosize & PAGE_MASK;
+}
+
 /*
  * Determine the maximum file size for a superblock
  */
@@ -775,17 +847,18 @@ void nfs_super_set_maxbytes(struct super_block *sb, __u64 maxfilesize)
  * Record the page as unstable (an extra writeback period) and mark its
  * inode as dirty.
  */
-static inline
-void nfs_mark_page_unstable(struct page *page, struct nfs_commit_info *cinfo)
+static inline void nfs_folio_mark_unstable(struct folio *folio,
+					   struct nfs_commit_info *cinfo)
 {
-	if (!cinfo->dreq) {
-		struct inode *inode = page_file_mapping(page)->host;
+	if (folio && !cinfo->dreq) {
+		struct inode *inode = folio->mapping->host;
+		long nr = folio_nr_pages(folio);
 
 		/* This page is really still in write-back - just that the
 		 * writeback is happening on the server now.
 		 */
-		inc_node_page_state(page, NR_WRITEBACK);
-		inc_wb_stat(&inode_to_bdi(inode)->wb, WB_WRITEBACK);
+		node_stat_mod_folio(folio, NR_WRITEBACK, nr);
+		wb_stat_mod(&inode_to_bdi(inode)->wb, WB_WRITEBACK, nr);
 		__mark_inode_dirty(inode, I_DIRTY_DATASYNC);
 	}
 }
@@ -793,18 +866,17 @@ void nfs_mark_page_unstable(struct page *page, struct nfs_commit_info *cinfo)
 /*
  * Determine the number of bytes of data the page contains
  */
-static inline
-unsigned int nfs_page_length(struct page *page)
+static inline size_t nfs_folio_length(struct folio *folio)
 {
-	loff_t i_size = i_size_read(page_file_mapping(page)->host);
+	loff_t i_size = i_size_read(folio->mapping->host);
 
 	if (i_size > 0) {
-		pgoff_t index = page_index(page);
-		pgoff_t end_index = (i_size - 1) >> PAGE_SHIFT;
+		pgoff_t index = folio->index >> folio_order(folio);
+		pgoff_t end_index = (i_size - 1) >> folio_shift(folio);
 		if (index < end_index)
-			return PAGE_SIZE;
+			return folio_size(folio);
 		if (index == end_index)
-			return ((i_size - 1) & ~PAGE_MASK) + 1;
+			return offset_in_folio(folio, i_size - 1) + 1;
 	}
 	return 0;
 }
@@ -822,11 +894,10 @@ unsigned char nfs_umode_to_dtype(umode_t mode)
  * Determine the number of pages in an array of length 'len' and
  * with a base offset of 'base'
  */
-static inline
-unsigned int nfs_page_array_len(unsigned int base, size_t len)
+static inline unsigned int nfs_page_array_len(unsigned int base, size_t len)
 {
-	return ((unsigned long)len + (unsigned long)base +
-		PAGE_SIZE - 1) >> PAGE_SHIFT;
+	return ((unsigned long)len + (unsigned long)base + PAGE_SIZE - 1) >>
+	       PAGE_SHIFT;
 }
 
 /*
@@ -842,33 +913,11 @@ u64 nfs_timespec_to_change_attr(const struct timespec64 *ts)
 	return ((u64)ts->tv_sec << 30) + ts->tv_nsec;
 }
 
-#ifdef CONFIG_CRC32
-/**
- * nfs_fhandle_hash - calculate the crc32 hash for the filehandle
- * @fh - pointer to filehandle
- *
- * returns a crc32 hash for the filehandle that is compatible with
- * the one displayed by "wireshark".
- */
-static inline u32 nfs_fhandle_hash(const struct nfs_fh *fh)
-{
-	return ~crc32_le(0xFFFFFFFF, &fh->data[0], fh->size);
-}
 static inline u32 nfs_stateid_hash(const nfs4_stateid *stateid)
 {
 	return ~crc32_le(0xFFFFFFFF, &stateid->other[0],
 				NFS4_STATEID_OTHER_SIZE);
 }
-#else
-static inline u32 nfs_fhandle_hash(const struct nfs_fh *fh)
-{
-	return 0;
-}
-static inline u32 nfs_stateid_hash(nfs4_stateid *stateid)
-{
-	return 0;
-}
-#endif
 
 static inline bool nfs_error_is_fatal(int err)
 {
@@ -903,31 +952,52 @@ static inline bool nfs_error_is_fatal_on_server(int err)
 	return nfs_error_is_fatal(err);
 }
 
-static inline bool get_no_idle_xprt(const struct nfs_fs_context *ctx)
-{
-	if (ctx->local_ports && ctx->local_ports->nr > 1 &&
-	    ctx->nfs_server.protocol == XPRT_TRANSPORT_RDMA &&
-	    ctx->no_idle_xprt == -1)
-		return true;
-	else if (ctx->no_idle_xprt != -1)
-		return (bool) ctx->no_idle_xprt;
-	else
-		return false;
-}
-
 /*
  * Select between a default port value and a user-specified port value.
  * If a zero value is set, then autobind will be used.
  */
-static inline void nfs_set_port(struct sockaddr *sap, int *port,
+static inline void nfs_set_port(struct sockaddr_storage *sap, int *port,
 				const unsigned short default_port)
 {
 	if (*port == NFS_UNSPEC_PORT)
 		*port = default_port;
 
-	rpc_set_port(sap, *port);
+	rpc_set_port((struct sockaddr *)sap, *port);
 }
 
+struct nfs_direct_req {
+	struct kref		kref;		/* release manager */
+
+	/* I/O parameters */
+	struct nfs_open_context	*ctx;		/* file open context info */
+	struct nfs_lock_context *l_ctx;		/* Lock context info */
+	struct kiocb *		iocb;		/* controlling i/o request */
+	struct inode *		inode;		/* target file of i/o */
+
+	/* completion state */
+	atomic_t		io_count;	/* i/os we're waiting for */
+	spinlock_t		lock;		/* protect completion state */
+
+	loff_t			io_start;	/* Start offset for I/O */
+	ssize_t			count,		/* bytes actually processed */
+				max_count,	/* max expected count */
+				bytes_left,	/* bytes left to be sent */
+				error;		/* any reported error */
+	struct completion	completion;	/* wait for i/o completion */
+
+	/* commit state */
+	struct nfs_mds_commit_info mds_cinfo;	/* Storage for cinfo */
+	struct pnfs_ds_commit_info ds_cinfo;	/* Storage for cinfo */
+	struct work_struct	work;
+	int			flags;
+	/* for write */
+#define NFS_ODIRECT_DO_COMMIT		(1)	/* an unstable reply was received */
+#define NFS_ODIRECT_RESCHED_WRITES	(2)	/* write verification failed */
+	/* for read */
+#define NFS_ODIRECT_SHOULD_DIRTY	(3)	/* dirty user-space page after read */
+#define NFS_ODIRECT_DONE		INT_MAX	/* write verification failed */
+};
+
 extern uint32_t nfs_metadata_hashing_mask;
 extern uint32_t nfs_metadata_client_mask;
 
@@ -946,7 +1016,7 @@ extern uint32_t nfs_metadata_client_mask;
 
 #define NFS_FH_HASH_INODE(inode)  \
 	NFS_MULTIPATH_HASH((nfs_fhandle_hash(NFS_FH(inode)) & nfs_metadata_hashing_mask) + \
-	  _NFS_FH_CLIENT_HASH(NFS_CLIENT(inode))) \
+	  _NFS_FH_CLIENT_HASH(NFS_CLIENT(inode)))
 
 extern struct nfs_server *nfs_multipath_get_client_by_dentry(struct dentry *dentry);
 
