@@ -11,20 +11,25 @@
  * nfs3acl.c
  */
 #ifdef CONFIG_NFS_V3_ACL
-#if defined(COMPAT_DETECT_GET_ACL_RCU_PARAM)
-extern struct posix_acl *nfs3_get_acl(struct inode *inode, int type, bool rcu);
-#else
+#ifdef COMPAT_DETECT_GET_ACL_OLD_SIGNATURE_5_15
 extern struct posix_acl *nfs3_get_acl(struct inode *inode, int type);
+#else
+extern struct posix_acl *nfs3_get_acl(struct inode *inode, int type, bool rcu);
 #endif
-extern int nfs3_set_acl(COMPAT_INIT_USER_NS_PARAM_N(mnt_userns) struct SET_ACL_PARAM_TYPE *param,
+#ifdef COMPAT_DETECT_SET_POSIX_ACL_USES_DENTRY_6_2
+extern int nfs3_set_acl(COMPAT_STRUCT_MNT_IDMAP *idmap, struct dentry *dentry,
 			struct posix_acl *acl, int type);
+#else
+extern int nfs3_set_acl(COMPAT_STRUCT_MNT_IDMAP *idmap, struct inode *inode,
+			struct posix_acl *acl, int type);
+#endif
 extern int nfs3_proc_setacls(struct inode *inode, struct posix_acl *acl,
 		struct posix_acl *dfacl);
 extern ssize_t nfs3_listxattr(struct dentry *, char *, size_t);
-#if defined(COMPAT_DETECT_POSIX_ACL_ACCESS_XATTR_HANDLER)
+#ifdef COMPAT_DETECT_POSIX_ACL_ACCESS_XATTR_HANDLER_6_4
 extern const struct xattr_handler *nfs3_xattr_handlers[];
 #endif
-#else
+#else /* !CONFIG_NFS_V3_ACL */
 static inline int nfs3_proc_setacls(struct inode *inode, struct posix_acl *acl,
 		struct posix_acl *dfacl)
 {
