@@ -11,6 +11,7 @@
 #include <linux/nfs_fs.h>
 #include <linux/hashtable.h>
 #include <linux/refcount.h>
+#include <linux/list_lru.h>
 #include <uapi/linux/xattr.h>
 
 #include "nfs4_fs.h"
@@ -132,7 +133,7 @@ nfs4_xattr_entry_lru_add(struct nfs4_xattr_entry *entry)
 	lru = (entry->flags & NFS4_XATTR_ENTRY_EXTVAL) ?
 	    &nfs4_xattr_large_entry_lru : &nfs4_xattr_entry_lru;
 
-	return compat_list_lru_add(lru, &entry->lru);
+	return list_lru_add(lru, &entry->lru);
 }
 
 static bool
@@ -143,7 +144,7 @@ nfs4_xattr_entry_lru_del(struct nfs4_xattr_entry *entry)
 	lru = (entry->flags & NFS4_XATTR_ENTRY_EXTVAL) ?
 	    &nfs4_xattr_large_entry_lru : &nfs4_xattr_entry_lru;
 
-	return compat_list_lru_del(lru, &entry->lru);
+	return list_lru_del(lru, &entry->lru);
 }
 
 /*
@@ -349,7 +350,7 @@ nfs4_xattr_cache_unlink(struct inode *inode)
 
 	oldcache = nfsi->xattr_cache;
 	if (oldcache != NULL) {
-		compat_list_lru_del(&nfs4_xattr_cache_lru, &oldcache->lru);
+		list_lru_del(&nfs4_xattr_cache_lru, &oldcache->lru);
 		oldcache->inode = NULL;
 	}
 	nfsi->xattr_cache = NULL;
@@ -474,7 +475,7 @@ nfs4_xattr_get_cache(struct inode *inode, int add)
 			kref_get(&cache->ref);
 			nfsi->xattr_cache = cache;
 			cache->inode = inode;
-			compat_list_lru_add(&nfs4_xattr_cache_lru, &cache->lru);
+			list_lru_add(&nfs4_xattr_cache_lru, &cache->lru);
 		}
 
 		spin_unlock(&inode->i_lock);
@@ -830,7 +831,11 @@ static struct shrinker nfs4_xattr_large_entry_shrinker = {
 
 static enum lru_status
 cache_lru_isolate(struct list_head *item,
+#ifdef COMPAT_DETECT_LIST_LRU_WALK_CB_6_13
+	struct list_lru_one *lru, void *arg)
+#else /* !COMPAT_DETECT_LIST_LRU_WALK_CB_6_13 */
 	struct list_lru_one *lru, spinlock_t *lru_lock, void *arg)
+#endif /* COMPAT_DETECT_LIST_LRU_WALK_CB_6_13 */
 {
 	struct list_head *dispose = arg;
 	struct inode *inode;
@@ -895,7 +900,11 @@ nfs4_xattr_cache_count(struct shrinker *shrink, struct shrink_control *sc)
 
 static enum lru_status
 entry_lru_isolate(struct list_head *item,
+#ifdef COMPAT_DETECT_LIST_LRU_WALK_CB_6_13
+	struct list_lru_one *lru, void *arg)
+#else /* !COMPAT_DETECT_LIST_LRU_WALK_CB_6_13 */
 	struct list_lru_one *lru, spinlock_t *lru_lock, void *arg)
+#endif /* COMPAT_DETECT_LIST_LRU_WALK_CB_6_13 */
 {
 	struct list_head *dispose = arg;
 	struct nfs4_xattr_bucket *bucket;
@@ -952,11 +961,7 @@ nfs4_xattr_entry_scan(struct shrinker *shrink, struct shrink_control *sc)
 	struct nfs4_xattr_entry *entry;
 	struct list_lru *lru;
 
-#ifdef COMPAT_DETECT_DYNAMIC_SHRINKER_ALLOC
-	lru = (shrink == nfs4_xattr_large_entry_shrinker) ?
-#else
-	lru = (shrink == &nfs4_xattr_large_entry_shrinker) ?
-#endif
+	lru = (shrink == COMPAT_SHRINKER_ADDRESS(nfs4_xattr_large_entry_shrinker)) ?
 	    &nfs4_xattr_large_entry_lru : &nfs4_xattr_entry_lru;
 
 	freed = list_lru_shrink_walk(lru, sc, entry_lru_isolate, &dispose);
@@ -984,11 +989,7 @@ nfs4_xattr_entry_count(struct shrinker *shrink, struct shrink_control *sc)
 	unsigned long count;
 	struct list_lru *lru;
 
-#ifdef COMPAT_DETECT_DYNAMIC_SHRINKER_ALLOC
-	lru = (shrink == nfs4_xattr_large_entry_shrinker) ?
-#else
-	lru = (shrink == &nfs4_xattr_large_entry_shrinker) ?
-#endif
+	lru = (shrink == COMPAT_SHRINKER_ADDRESS(nfs4_xattr_large_entry_shrinker)) ?
 	    &nfs4_xattr_large_entry_lru : &nfs4_xattr_entry_lru;
 
 	count = list_lru_shrink_count(lru, sc);
@@ -998,7 +999,7 @@ nfs4_xattr_entry_count(struct shrinker *shrink, struct shrink_control *sc)
 
 static void nfs4_xattr_cache_init_once(void *p)
 {
-	struct nfs4_xattr_cache *cache = (struct nfs4_xattr_cache *)p;
+	struct nfs4_xattr_cache *cache = p;
 
 	spin_lock_init(&cache->listxattr_lock);
 	atomic_long_set(&cache->nent, 0);
@@ -1013,14 +1014,11 @@ typedef unsigned long (*count_objects_cb)(struct shrinker *s,
 typedef unsigned long (*scan_objects_cb)(struct shrinker *s,
 					 struct shrink_control *sc);
 
-#ifdef COMPAT_DETECT_DYNAMIC_SHRINKER_ALLOC
-static int __init nfs4_xattr_shrinker_init(struct shrinker **shrinker,
-#else
-static int __init nfs4_xattr_shrinker_init(struct shrinker *shrinker,
-#endif
-					   struct list_lru *lru, const char *name,
-					   count_objects_cb count,
-					   scan_objects_cb scan, long batch, int seeks)
+static int __init nfs4_xattr_shrinker_init(COMPAT_DYNAMIC_SHRINKER_TYPE *shrinker,
+                                           struct list_lru *lru, const char *name,
+                                           count_objects_cb count,
+                                           scan_objects_cb scan, long batch, int seeks)
+
 {
 	int ret;
 #ifdef COMPAT_DETECT_DYNAMIC_SHRINKER_ALLOC
@@ -1040,28 +1038,24 @@ static int __init nfs4_xattr_shrinker_init(struct shrinker *shrinker,
 	(*shrinker)->seeks = seeks;
 
 	shrinker_register(*shrinker);
-#else
-	ret = compat_list_lru_init_memcg(lru, shrinker);
+#else /* !COMPAT_DETECT_DYNAMIC_SHRINKER_ALLOC */
+	ret = register_shrinker(shrinker, name);
 	if (ret)
 		return ret;
 
-	ret = register_shrinker(shrinker, name);
+	ret = list_lru_init_memcg(lru, shrinker);
 	if (ret)
-		list_lru_destroy(lru);
-#endif
+		unregister_shrinker(shrinker);
+#endif /* COMPAT_DETECT_DYNAMIC_SHRINKER_ALLOC */
 
 	return ret;
 }
-#ifdef COMPAT_DETECT_DYNAMIC_SHRINKER_ALLOC
-static void nfs4_xattr_shrinker_destroy(struct shrinker **shrinker,
-#else
+
 static void nfs4_xattr_shrinker_destroy(struct shrinker *shrinker,
-#endif
 					struct list_lru *lru)
 {
 #ifdef COMPAT_DETECT_DYNAMIC_SHRINKER_ALLOC
-	shrinker_free(*shrinker);
-	*shrinker = NULL;
+	shrinker_free(shrinker);
 #else
 	unregister_shrinker(shrinker);
 #endif
@@ -1101,10 +1095,10 @@ int __init nfs4_xattr_cache_init(void)
 	if (!ret)
 		return 0;
 
-	nfs4_xattr_shrinker_destroy(&nfs4_xattr_entry_shrinker,
+	nfs4_xattr_shrinker_destroy(COMPAT_SHRINKER_ADDRESS(nfs4_xattr_entry_shrinker),
 				    &nfs4_xattr_entry_lru);
 out2:
-	nfs4_xattr_shrinker_destroy(&nfs4_xattr_cache_shrinker,
+	nfs4_xattr_shrinker_destroy(COMPAT_SHRINKER_ADDRESS(nfs4_xattr_cache_shrinker),
 				    &nfs4_xattr_cache_lru);
 out1:
 	kmem_cache_destroy(nfs4_xattr_cache_cachep);
@@ -1114,11 +1108,11 @@ out1:
 
 void nfs4_xattr_cache_exit(void)
 {
-	nfs4_xattr_shrinker_destroy(&nfs4_xattr_large_entry_shrinker,
+	nfs4_xattr_shrinker_destroy(COMPAT_SHRINKER_ADDRESS(nfs4_xattr_large_entry_shrinker),
 				    &nfs4_xattr_large_entry_lru);
-	nfs4_xattr_shrinker_destroy(&nfs4_xattr_entry_shrinker,
+	nfs4_xattr_shrinker_destroy(COMPAT_SHRINKER_ADDRESS(nfs4_xattr_entry_shrinker),
 				    &nfs4_xattr_entry_lru);
-	nfs4_xattr_shrinker_destroy(&nfs4_xattr_cache_shrinker,
+	nfs4_xattr_shrinker_destroy(COMPAT_SHRINKER_ADDRESS(nfs4_xattr_cache_shrinker),
 				    &nfs4_xattr_cache_lru);
 	kmem_cache_destroy(nfs4_xattr_cache_cachep);
 }
