@@ -48,6 +48,7 @@
 #include <linux/vfs.h>
 #include <linux/inet.h>
 #include <linux/in6.h>
+#include <linux/sched.h>
 #include <linux/slab.h>
 #include <net/ipv6.h>
 #include <linux/netdevice.h>
@@ -60,6 +61,8 @@
 #include <linux/uaccess.h>
 #include <linux/nfs_ssc.h>
 
+#include <uapi/linux/tls.h>
+
 #include "nfs4_fs.h"
 #include "callback.h"
 #include "delegation.h"
@@ -70,16 +73,14 @@
 #include "nfstrace.h"
 #include "pnfs.h"
 #include "nfs.h"
+#include "netns.h"
+#include "sysfs.h"
 
 #define NFSDBG_FACILITY		NFSDBG_VFS
 
 const struct super_operations nfs_sops = {
 	.alloc_inode	= nfs_alloc_inode,
-#ifdef COMPAT_DETECT_FREE_INODE
 	.free_inode	= nfs_free_inode,
-#else
-	.destroy_inode	= nfs_destroy_inode,
-#endif
 	.write_inode	= nfs_write_inode,
 	.drop_inode	= nfs_drop_inode,
 	.statfs		= nfs_statfs,
@@ -101,12 +102,12 @@ static const struct nfs_ssc_client_ops nfs_ssc_clnt_ops_tbl = {
 #if IS_ENABLED(CONFIG_NFS_V4)
 static int __init register_nfs4_fs(void)
 {
-	return COMPAT_REGISTER_FILESYSTEM(nfs4_fs_type);
+	return register_filesystem(&nfs4_fs_type);
 }
 
 static void unregister_nfs4_fs(void)
 {
-	COMPAT_UNREGISTER_FILESYSTEM(nfs4_fs_type);
+	unregister_filesystem(&nfs4_fs_type);
 }
 #else
 static int __init register_nfs4_fs(void)
@@ -148,7 +149,7 @@ int __init register_nfs_fs(void)
 {
 	int ret;
 
-        ret = COMPAT_REGISTER_FILESYSTEM(nfs_fs_type);
+	ret = register_filesystem(&nfs_fs_type);
 	if (ret < 0)
 		goto error_0;
 
@@ -159,7 +160,6 @@ int __init register_nfs_fs(void)
 	ret = nfs_register_sysctl();
 	if (ret < 0)
 		goto error_2;
-
 #ifdef COMPAT_DETECT_DYNAMIC_SHRINKER_ALLOC
 	acl_shrinker = shrinker_alloc(0, "nfs-acl");
 	if (!acl_shrinker) {
@@ -171,11 +171,11 @@ int __init register_nfs_fs(void)
 	acl_shrinker->scan_objects = nfs_access_cache_scan;
 
 	shrinker_register(acl_shrinker);
-#else
+#else /* !COMPAT_DETECT_DYNAMIC_SHRINKER_ALLOC */
 	ret = register_shrinker(&acl_shrinker, "nfs-acl");
 	if (ret < 0)
 		goto error_3;
-#endif
+#endif /* COMPAT_DETECT_DYNAMIC_SHRINKER_ALLOC */
 #ifdef CONFIG_NFS_V4_2
 	nfs_ssc_register_ops();
 #endif
@@ -185,7 +185,7 @@ error_3:
 error_2:
 	unregister_nfs4_fs();
 error_1:
-	COMPAT_UNREGISTER_FILESYSTEM(nfs_fs_type);
+	unregister_filesystem(&nfs_fs_type);
 error_0:
 	return ret;
 }
@@ -195,17 +195,13 @@ error_0:
  */
 void __exit unregister_nfs_fs(void)
 {
-#ifdef COMPAT_DETECT_DYNAMIC_SHRINKER_ALLOC
-	shrinker_free(acl_shrinker);
-#else
-	unregister_shrinker(&acl_shrinker);
-#endif
+	compat_shrinker_free(&acl_shrinker);
 	nfs_unregister_sysctl();
 	unregister_nfs4_fs();
 #ifdef CONFIG_NFS_V4_2
 	nfs_ssc_unregister_ops();
 #endif
-	COMPAT_UNREGISTER_FILESYSTEM(nfs_fs_type);
+	unregister_filesystem(&nfs_fs_type);
 }
 
 bool nfs_sb_active(struct super_block *sb)
@@ -247,6 +243,7 @@ static int __nfs_list_for_each_server(struct list_head *head,
 		ret = fn(server, data);
 		if (ret)
 			goto out;
+		cond_resched();
 		rcu_read_lock();
 	}
 	rcu_read_unlock();
@@ -278,7 +275,6 @@ int nfs_statfs(struct dentry *dentry, struct kstatfs *buf)
 	int error = -ENOMEM;
 
 	trace_nfs_statfs_enter(inode);
-
 	res.fattr = nfs_alloc_fattr();
 	if (res.fattr == NULL)
 		goto out_err;
@@ -481,7 +477,7 @@ int nfs_get_succeeding_ipaddr(const struct sockaddr *addr,
 
 	memset(buff, 0, sizeof(*buff));
 	switch (addr->sa_family) {
-	case AF_INET: 
+	case AF_INET:
 		in = (struct sockaddr_in *)buff;
 		memcpy(in, addr, sizeof(*in));
 
@@ -494,7 +490,7 @@ int nfs_get_succeeding_ipaddr(const struct sockaddr *addr,
 
 		if (!(in6->sin6_addr.s6_addr32[3] =
 		htonl(ntohl(in6->sin6_addr.s6_addr32[3]) + 1))) /* Overflow */
-			in6->sin6_addr.s6_addr32[2] = 
+			in6->sin6_addr.s6_addr32[2] =
 				htonl(ntohl(in6->sin6_addr.s6_addr32[2]) + 1);
 		retval = sizeof(*in6);
 		break;
@@ -577,10 +573,11 @@ static void nfs_show_mount_options(struct seq_file *m, struct nfs_server *nfss,
 		{ NFS_MOUNT_NONLM, ",nolock", "" },
 		{ NFS_MOUNT_NOACL, ",noacl", "" },
 		{ NFS_MOUNT_NORDIRPLUS, ",nordirplus", "" },
+		{ NFS_MOUNT_FORCE_RDIRPLUS, ",rdirplus=force", "" },
 		{ NFS_MOUNT_UNSHARED, ",nosharecache", "" },
 		{ NFS_MOUNT_NOSHARE_XPRT, ",nosharetransport", ""},
 		{ NFS_MOUNT_NORESVPORT, ",noresvport", "" },
-		{ NFS_MOUNT_FORCE_READDIRPLUS, ",forcerdirplus", "" },
+		{ NFS_MOUNT_FORCE_RDIRPLUS, ",forcerdirplus", "" },
 		{ NFS_MOUNT_OPT_LOCK_FLUSH, ",optlockflush", "" },
 		{ NFS_MOUNT_RELMTIME, ",relmtime", "" },
 		{ NFS_MOUNT_NO_EXTEND, ",noextend", "" },
@@ -632,6 +629,16 @@ static void nfs_show_mount_options(struct seq_file *m, struct nfs_server *nfss,
 	seq_printf(m, ",timeo=%lu", 10U * nfss->client->cl_timeout->to_initval / HZ);
 	seq_printf(m, ",retrans=%u", nfss->client->cl_timeout->to_retries);
 	seq_printf(m, ",sec=%s", nfs_pseudoflavour_to_name(nfss->client->cl_auth->au_flavor));
+	switch (clp->cl_xprtsec.policy) {
+	case RPC_XPRTSEC_TLS_ANON:
+		seq_puts(m, ",xprtsec=tls");
+		break;
+	case RPC_XPRTSEC_TLS_X509:
+		seq_puts(m, ",xprtsec=mtls");
+		break;
+	default:
+		break;
+	}
 
 	if (version != 4)
 		nfs_show_mountd_options(m, nfss, showdefaults);
@@ -686,8 +693,6 @@ static void nfs_show_mount_options(struct seq_file *m, struct nfs_server *nfss,
                 seq_puts(m, ",noidlexprt");
 	if (clp->cl_mdconnect)
                 seq_printf(m, ",mdconnect=%d", clp->cl_mdconnect);
-	if (clp->cl_pconnect >= 2)
-                seq_printf(m, ",pconnect=%d", clp->cl_pconnect);
 	if (clp->cl_spread_reads)
 		seq_puts(m, ",spread_reads");
 	if (clp->cl_spread_writes)
@@ -856,10 +861,6 @@ int nfs_show_stats(struct seq_file *m, struct dentry *root)
 			totals.events[i] += stats->events[i];
 		for (i = 0; i < __NFSIOS_BYTESMAX; i++)
 			totals.bytes[i] += stats->bytes[i];
-#ifdef CONFIG_NFS_FSCACHE
-		for (i = 0; i < __NFSIOS_FSCACHEMAX; i++)
-			totals.fscache[i] += stats->fscache[i];
-#endif
 
 		preempt_enable();
 	}
@@ -870,13 +871,6 @@ int nfs_show_stats(struct seq_file *m, struct dentry *root)
 	seq_puts(m, "\n\tbytes:\t");
 	for (i = 0; i < __NFSIOS_BYTESMAX; i++)
 		seq_printf(m, "%Lu ", totals.bytes[i]);
-#ifdef CONFIG_NFS_FSCACHE
-	if (nfss->options & NFS_OPTION_FSCACHE) {
-		seq_puts(m, "\n\tfsc:\t");
-		for (i = 0; i < __NFSIOS_FSCACHEMAX; i++)
-			seq_printf(m, "%Lu ", totals.fscache[i]);
-	}
-#endif
 	seq_putc(m, '\n');
 
 	rpc_clnt_show_stats(m, nfss->client, false);
@@ -896,7 +890,6 @@ int nfs_show_stats(struct seq_file *m, struct dentry *root)
 		if (found > 0)
 			seq_printf(m, "\n");
 	}
-
 	return 0;
 }
 EXPORT_SYMBOL_GPL(nfs_show_stats);
@@ -1002,8 +995,7 @@ static int nfs_request_mount(struct fs_context *fc,
 {
 	struct nfs_fs_context *ctx = nfs_fc2context(fc);
 	struct nfs_mount_request request = {
-		.sap		= (struct sockaddr *)
-						&ctx->mount_server.address,
+		.sap		= &ctx->mount_server._address,
 		.dirpath	= ctx->nfs_server.export_path,
 		.protocol	= ctx->mount_server.protocol,
 		.fh		= root_fh,
@@ -1034,7 +1026,7 @@ static int nfs_request_mount(struct fs_context *fc,
 	 * Construct the mount server's address.
 	 */
 	if (ctx->mount_server.address.sa_family == AF_UNSPEC) {
-		memcpy(request.sap, &ctx->nfs_server.address,
+		memcpy(request.sap, &ctx->nfs_server._address,
 		       ctx->nfs_server.addrlen);
 		ctx->mount_server.addrlen = ctx->nfs_server.addrlen;
 	}
@@ -1045,7 +1037,15 @@ static int nfs_request_mount(struct fs_context *fc,
 	 * Now ask the mount server to map our export path
 	 * to a file handle.
 	 */
-	status = nfs_mount(&request, ctx->timeo, ctx->retrans);
+	if ((request.protocol == XPRT_TRANSPORT_UDP) ==
+	    !(ctx->flags & NFS_MOUNT_TCP))
+		/*
+		 * NFS protocol and mount protocol are both UDP or neither UDP
+		 * so timeouts are compatible.  Use NFS timeouts for MOUNT
+		 */
+		status = nfs_mount(&request, ctx->timeo, ctx->retrans);
+	else
+		status = nfs_mount(&request, NFS_UNSPEC_TIMEO, NFS_UNSPEC_RETRANS);
 	if (status != 0) {
 		dfprintk(MOUNT, "NFS: unable to mount server %s, error %d\n",
 				request.hostname, status);
@@ -1184,6 +1184,7 @@ int nfs_reconfigure(struct fs_context *fc)
 	struct nfs_fs_context *ctx = nfs_fc2context(fc);
 	struct super_block *sb = fc->root->d_sb;
 	struct nfs_server *nfss = sb->s_fs_info;
+	int ret;
 
 	sync_filesystem(sb);
 
@@ -1208,7 +1209,11 @@ int nfs_reconfigure(struct fs_context *fc)
 	}
 
 	/* compare new mount options with old ones */
-	return nfs_compare_remount_data(nfss, ctx);
+	ret = nfs_compare_remount_data(nfss, ctx);
+	if (ret)
+		return ret;
+
+	return nfs_probe_server(nfss, NFS_FH(d_inode(fc->root)));
 }
 EXPORT_SYMBOL_GPL(nfs_reconfigure);
 
@@ -1229,10 +1234,8 @@ static void nfs_fill_super(struct super_block *sb, struct nfs_fs_context *ctx)
 	switch (server->nfs_client->rpc_ops->version) {
 	case 2:
 		sb->s_time_gran = 1000;
-#if defined(COMPAT_DETECT_FS_TIMESTAMP_RANGES)
 		sb->s_time_min = 0;
 		sb->s_time_max = U32_MAX;
-#endif
 		break;
 	case 3:
 		/*
@@ -1241,19 +1244,15 @@ static void nfs_fill_super(struct super_block *sb, struct nfs_fs_context *ctx)
 		 */
 		sb->s_flags |= SB_POSIXACL;
 		sb->s_time_gran = 1;
-#if defined(COMPAT_DETECT_FS_TIMESTAMP_RANGES)
 		sb->s_time_min = 0;
 		sb->s_time_max = U32_MAX;
-#endif
 		sb->s_export_op = &nfs_export_ops;
 		break;
 	case 4:
 		sb->s_flags |= SB_POSIXACL;
 		sb->s_time_gran = 1;
-#if defined(COMPAT_DETECT_FS_TIMESTAMP_RANGES)
 		sb->s_time_min = S64_MIN;
 		sb->s_time_max = S64_MAX;
-#endif
 		if (server->caps & NFS_CAP_ATOMIC_OPEN_V1)
 			sb->s_export_op = &nfs_export_ops;
 		break;
@@ -1270,6 +1269,7 @@ static void nfs_fill_super(struct super_block *sb, struct nfs_fs_context *ctx)
 						 &sb->s_blocksize_bits);
 
 	nfs_super_set_maxbytes(sb, server->maxfilesize);
+	nfs_sysfs_move_server_to_sb(sb);
 	server->has_sec_mnt_opts = ctx->has_sec_mnt_opts;
 }
 
@@ -1310,7 +1310,8 @@ static int nfs_set_super(struct super_block *s, struct fs_context *fc)
 	struct nfs_server *server = fc->s_fs_info;
 	int ret;
 
-	s->s_d_op = server->nfs_client->rpc_ops->dentry_ops;
+	set_default_d_op(s, server->nfs_client->rpc_ops->dentry_ops);
+
 	ret = set_anon_super(s, server);
 	if (ret == 0)
 		server->s_dev = s->s_dev;
@@ -1389,65 +1390,57 @@ static int nfs_compare_super(struct super_block *sb, struct fs_context *fc)
 		return 0;
 	if (!nfs_compare_userns(old, server))
 		return 0;
-#ifdef COMPAT_DETECT_SECURITY_SB_MNT_OPTS_COMPAT
 	if ((old->has_sec_mnt_opts || fc->security) &&
 			security_sb_mnt_opts_compat(sb, fc->security))
 		return 0;
-#endif
 	return nfs_compare_mount_options(sb, server, fc);
 }
 
 #ifdef CONFIG_NFS_FSCACHE
-static void nfs_get_cache_cookie(struct super_block *sb,
-				 struct nfs_fs_context *ctx)
+static int nfs_get_cache_cookie(struct super_block *sb,
+				struct nfs_fs_context *ctx)
 {
 	struct nfs_server *nfss = NFS_SB(sb);
 	char *uniq = NULL;
 	int ulen = 0;
 
-	nfss->fscache_key = NULL;
 	nfss->fscache = NULL;
 
 	if (!ctx)
-		return;
+		return 0;
 
 	if (ctx->clone_data.sb) {
 		struct nfs_server *mnt_s = NFS_SB(ctx->clone_data.sb);
 		if (!(mnt_s->options & NFS_OPTION_FSCACHE))
-			return;
-		if (mnt_s->fscache_key) {
-			uniq = mnt_s->fscache_key->key.uniquifier;
-			ulen = mnt_s->fscache_key->key.uniq_len;
+			return 0;
+		if (mnt_s->fscache_uniq) {
+			uniq = mnt_s->fscache_uniq;
+			ulen = strlen(uniq);
 		}
 	} else {
 		if (!(ctx->options & NFS_OPTION_FSCACHE))
-			return;
+			return 0;
 		if (ctx->fscache_uniq) {
 			uniq = ctx->fscache_uniq;
 			ulen = strlen(ctx->fscache_uniq);
 		}
 	}
 
-	nfs_fscache_get_super_cookie(sb, uniq, ulen);
+	return nfs_fscache_get_super_cookie(sb, uniq, ulen);
 }
 #else
-static void nfs_get_cache_cookie(struct super_block *sb,
-				 struct nfs_fs_context *ctx)
+static int nfs_get_cache_cookie(struct super_block *sb,
+				struct nfs_fs_context *ctx)
 {
+	return 0;
 }
 #endif
 
 static void nfs_set_readahead(struct backing_dev_info *bdi,
 			      struct nfs_server *server)
 {
-#if defined(VM_READAHEAD_PAGES)
 	bdi->ra_pages = VM_READAHEAD_PAGES;
 	bdi->io_pages = server->rpages;
-#else
-	/* Before commit - c128e575514ce9: NFS: Optimise the default readahead size */
-#define NFS_MAX_READAHEAD      (RPC_DEF_SLOT_TABLE - 1)
-	bdi->ra_pages = server->rpages * NFS_MAX_READAHEAD;
-#endif
 }
 
 int nfs_get_tree_common(struct fs_context *fc)
@@ -1456,7 +1449,7 @@ int nfs_get_tree_common(struct fs_context *fc)
 	struct super_block *s;
 	int (*compare_super)(struct super_block *, struct fs_context *) = nfs_compare_super;
 	struct nfs_server *server = ctx->server, *alt_server;
-	int error;
+	int error, i;
 
 	ctx->server = NULL;
 	if (IS_ERR(server))
@@ -1473,11 +1466,6 @@ int nfs_get_tree_common(struct fs_context *fc)
 		if (ctx->clone_data.sb->s_flags & SB_SYNCHRONOUS)
 			fc->sb_flags |= SB_SYNCHRONOUS;
 
-#if defined(COMPAT_DETECT_FS_CONTEXT_LSM_FLAGS)
-	if (server->caps & NFS_CAP_SECURITY_LABEL)
-		fc->lsm_flags |= SECURITY_LSM_NATIVE_LABELS;
-#endif
-
 	/* Get a superblock - note that we may end up sharing one that already exists */
 	fc->s_fs_info = server;
 	s = sget_fc(fc, compare_super, nfs_set_super);
@@ -1492,8 +1480,6 @@ int nfs_get_tree_common(struct fs_context *fc)
 		nfs_free_server(server);
 		server = NULL;
 	} else {
-		int i;
-
 		error = super_setup_bdi_name(s, "%u:%u", MAJOR(server->s_dev),
 					     MINOR(server->s_dev));
 		if (error)
@@ -1527,7 +1513,9 @@ int nfs_get_tree_common(struct fs_context *fc)
 			s->s_blocksize_bits = bsize;
 			s->s_blocksize = 1U << bsize;
 		}
-		nfs_get_cache_cookie(s, ctx);
+		error = nfs_get_cache_cookie(s, ctx);
+		if (error < 0)
+			goto error_splat_super;
 	}
 
 	error = nfs_get_root(s, fc);
@@ -1551,19 +1539,18 @@ error_splat_super:
 }
 
 /*
- * Destroy an NFS2/3 superblock
+ * Destroy an NFS superblock
  */
 void nfs_kill_super(struct super_block *s)
 {
 	struct nfs_server *server = NFS_SB(s);
-	dev_t dev = s->s_dev;
 
-	generic_shutdown_super(s);
+	nfs_sysfs_move_sb_to_server(server);
+	kill_anon_super(s);
 
 	nfs_fscache_release_super_cookie(s);
 
 	nfs_free_server(server);
-	free_anon_bdev(dev);
 }
 EXPORT_SYMBOL_GPL(nfs_kill_super);
 
