@@ -75,13 +75,13 @@ struct nfs_client {
 	u32			cl_minorversion;/* NFSv4 minorversion */
 	unsigned int		cl_nconnect;	/* Number of connections */
 	unsigned int		cl_max_connect; /* max number of xprts allowed */
-	const char *		cl_principal;  /* used for machine cred */
-	unsigned int		cl_pconnect;    /* number of connections per remote port */
+	const char *		cl_principal;	/* used for machine cred */
+	struct xprtsec_parms	cl_xprtsec;	/* xprt security policy */
 
 	unsigned int		cl_xprt_id;
 #if IS_ENABLED(CONFIG_NFS_V4)
 	unsigned int		cl_server_idx;	/* NFSv4.1 multipath server index */
-	struct list_head	cl_ds_clients; /* auth flavor data servers */
+	struct list_head	cl_ds_clients;	/* auth flavor data servers */
 	u64			cl_clientid;	/* constant */
 	nfs4_verifier		cl_confirm;	/* Clientid verifier */
 	unsigned long		cl_state;
@@ -135,18 +135,14 @@ struct nfs_client {
 	 * This is used to generate the mv0 callback address.
 	 */
 	char			cl_ipaddr[48];
-
-#ifdef CONFIG_NFS_FSCACHE
-	struct fscache_cookie	*fscache;	/* client index cache cookie */
-#endif
 	struct dentry		*cl_debugfs;
 
 	struct net		*cl_net;
 	struct list_head	pending_cb_stateids;
 };
 
-extern void nfs_client_debugfs_register(struct nfs_client* clp);
-extern void nfs_client_debugfs_unregister(struct nfs_client* clp);
+void nfs_client_debugfs_register(struct nfs_client* clp);
+void nfs_client_debugfs_unregister(struct nfs_client* clp);
 
 #define NFS_MAX_SERVERS 64
 
@@ -168,7 +164,7 @@ struct nfs_superblock {
  * NFS client parameters stored in the superblock.
  */
 struct nfs_server {
-	struct nfs_superblock   superblock;
+	struct nfs_superblock	superblock;
 	struct nfs_client *	nfs_client;	/* shared client and NFS4 state */
 	struct list_head	client_link;	/* List of other nfs_server structs
 						 * that share the same client
@@ -179,7 +175,7 @@ struct nfs_server {
 	struct nlm_host		*nlm_host;	/* NLM client handle */
 	struct nfs_iostats __percpu *io_stats;	/* I/O statistics */
 	atomic_long_t		writeback;	/* number of writeback pages */
-	unsigned int            write_congested;/* flag set when writeback gets too high */
+	unsigned int		write_congested;/* flag set when writeback gets too high */
 	unsigned int		flags;		/* various flags */
 	unsigned int		idx;		/* NFSv4.1 multipath server index */
 
@@ -194,11 +190,13 @@ struct nfs_server {
 #define NFS_MOUNT_SOFTREVAL		0x800000
 #define NFS_MOUNT_WRITE_EAGER		0x01000000
 #define NFS_MOUNT_WRITE_WAIT		0x02000000
-#define NFS_MOUNT_NO_EXTEND		0x04000000
-#define NFS_MOUNT_FORCE_READDIRPLUS	0x08000000
-#define NFS_MOUNT_OPT_LOCK_FLUSH	0x10000000
-#define NFS_MOUNT_RELMTIME	        0x20000000
-#define NFS_MOUNT_NOSHARE_XPRT		0x40000000
+#define NFS_MOUNT_TRUNK_DISCOVERY	0x04000000
+#define NFS_MOUNT_SHUTDOWN		0x08000000
+#define NFS_MOUNT_NO_EXTEND		0x10000000
+#define NFS_MOUNT_FORCE_RDIRPLUS	0x20000000
+#define NFS_MOUNT_OPT_LOCK_FLUSH	0x40000000
+#define NFS_MOUNT_RELMTIME	        0x80000000
+#define NFS_MOUNT_NOSHARE_XPRT		0xc0000000
 
 	unsigned int		fattr_valid;	/* Valid attributes */
 	unsigned int		caps;		/* server capabilities */
@@ -229,6 +227,7 @@ struct nfs_server {
 				change_attr_type;/* Description of change attribute */
 
 	struct nfs_fsid		fsid;
+	int			s_sysfs_id;	/* sysfs dentry index */
 	__u64			maxfilesize;	/* maximum file size */
 	struct timespec64	time_delta;	/* smallest time granularity */
 	unsigned long		mount_time;	/* when this fs was mounted */
@@ -237,8 +236,8 @@ struct nfs_server {
 	struct nfs_auth_info	auth_info;	/* parsed auth flavors */
 
 #ifdef CONFIG_NFS_FSCACHE
-	struct nfs_fscache_key	*fscache_key;	/* unique key for superblock */
-	struct fscache_cookie	*fscache;	/* superblock cookie */
+	struct fscache_volume	*fscache;	/* superblock cookie */
+	char			*fscache_uniq;	/* Uniquifier (or NULL) */
 #endif
 
 	u32			pnfs_blksize;	/* layout_blksize attr */
@@ -280,7 +279,13 @@ struct nfs_server {
 	struct list_head	layouts;
 	struct list_head	delegations;
 	struct list_head	ss_copies;
+	struct list_head	ss_src_copies;
 
+	unsigned long		delegation_flags;
+#define NFS4SERV_DELEGRETURN		(1)
+#define NFS4SERV_DELEGATION_EXPIRED	(2)
+#define NFS4SERV_DELEGRETURN_DELAYED	(3)
+	unsigned long		delegation_gen;
 	unsigned long		mig_gen;
 	unsigned long		mig_status;
 #define NFS_MIG_IN_TRANSITION		(1)
@@ -302,17 +307,15 @@ struct nfs_server {
 	/* XDR related information */
 	unsigned int		read_hdrsize;
 
-	/* ID for tracing */
-	unsigned int		trid;
-
 	/* User namespace info */
 	const struct cred	*cred;
 
 	/* Saved remote ports for NFSv4.x */
 	struct rpc_portgroup    *remote_ports;
-
 	struct dentry		*debugfs;
+
 	bool			has_sec_mnt_opts;
+	struct kobject		kobj;
 };
 
 /* Server capabilities */
@@ -322,6 +325,10 @@ struct nfs_server {
 #define NFS_CAP_ACLS		(1U << 3)
 #define NFS_CAP_ATOMIC_OPEN	(1U << 4)
 #define NFS_CAP_LGOPEN		(1U << 5)
+#define NFS_CAP_CASE_INSENSITIVE	(1U << 6)
+#define NFS_CAP_CASE_PRESERVING	(1U << 7)
+#define NFS_CAP_OPEN_XOR	(1U << 12)
+#define NFS_CAP_DELEGTIME	(1U << 13)
 #define NFS_CAP_POSIX_LOCK	(1U << 14)
 #define NFS_CAP_UIDGID_NOMAP	(1U << 15)
 #define NFS_CAP_STATEID_NFSV41	(1U << 16)
@@ -341,7 +348,7 @@ struct nfs_server {
 #define NFS_CAP_FS_LOCATIONS	(1U << 30)
 #define NFS_CAP_MOVEABLE	(1U << 31)
 
-extern void nfs_server_debugfs_register(struct nfs_server *server);
-extern void nfs_server_debugfs_unregister(struct nfs_server *server);
+void nfs_server_debugfs_register(struct nfs_server *server);
+void nfs_server_debugfs_unregister(struct nfs_server *server);
 
 #endif
