@@ -4,9 +4,10 @@
 /*
  * v6.8-rc1~180^2~315 "fs: convert error_remove_page to error_remove_folio"
  */
-const void *check_func(struct address_space_operations *aops)
+static const void *check_func(struct address_space_operations *aops)
 {
 	return aops->error_remove_folio;
 }
 
+EXPORT_SYMBOL(check_func);
 MODULE_LICENSE("GPL");
