@@ -3,9 +3,10 @@
 
 /* See commit v6.1-rc1~98^2~6 */
 
-int check_type(void)
+static int check_type(void)
 {
 	return TASK_FREEZABLE;
 }
 
+EXPORT_SYMBOL(check_type);
 MODULE_LICENSE("GPL");
