@@ -52,7 +52,7 @@
 #include <linux/sunrpc/addr.h>
 #include <linux/sunrpc/xprt.h>
 #include <linux/rcupdate.h>
-#include <linux/sched.h>
+#include <linux/sched/mm.h>
 
 #include <trace/events/sunrpc.h>
 
@@ -820,8 +820,6 @@ xprt_schedule_autodisconnect(struct rpc_xprt *xprt)
 		mod_timer(&xprt->timer, xprt->last_used + xprt->idle_timeout);
 }
 
-#include "timer.h"
-
 static void
 xprt_init_autodisconnect(struct timer_list *t)
 {
@@ -1048,8 +1046,6 @@ xprt_is_pinned_rqst(struct rpc_rqst *req)
 	return atomic_read(&req->rq_pin) != 0;
 }
 
-#include "wake_up_var.h"
-
 /**
  * xprt_pin_rqst - Pin a request on the transport receive list
  * @req: Request to pin
