@@ -13,6 +13,7 @@
 #include <linux/module.h>
 #include <linux/slab.h>
 #include <linux/errno.h>
+#include <linux/sort.h>
 #include <linux/hash.h>
 #include <linux/sunrpc/clnt.h>
 #include <linux/sunrpc/gss_api.h>
@@ -456,7 +457,7 @@ rpcauth_prune_expired(struct list_head *free, int nr_to_scan)
 		freed++;
 		list_add_tail(&cred->cr_lru, free);
 	}
-	return freed ? freed : -1;
+	return freed ? freed : SHRINK_STOP;
 }
 
 static unsigned long
@@ -478,24 +479,23 @@ rpcauth_cache_do_shrink(int nr_to_scan)
  */
 static unsigned long
 rpcauth_cache_shrink_scan(struct shrinker *shrink, struct shrink_control *sc)
-{
-	int nr_to_scan = sc->nr_to_scan;
-	gfp_t gfp_mask = sc->gfp_mask;
 
-	if ((gfp_mask & GFP_KERNEL) != GFP_KERNEL)
-		return (nr_to_scan == 0) ? 0 : -1;
+{
+	if ((sc->gfp_mask & GFP_KERNEL) != GFP_KERNEL)
+		return SHRINK_STOP;
 
 	/* nothing left, don't come back */
 	if (list_empty(&cred_unused))
-		return 0;
+		return SHRINK_STOP;
 
 	return rpcauth_cache_do_shrink(sc->nr_to_scan);
 }
 
-static int
-rpcauth_cache_shrinker(struct shrinker *shrink, struct shrink_control *sc)
+static unsigned long
+rpcauth_cache_shrink_count(struct shrinker *shrink, struct shrink_control *sc)
+
 {
-	return rpcauth_cache_shrink_scan(shrink, sc);
+	return number_cred_unused * sysctl_vfs_cache_pressure / 100;
 }
 
 static void
@@ -886,7 +886,8 @@ rpcauth_uptodatecred(struct rpc_task *task)
 }
 
 static struct shrinker rpc_cred_shrinker = {
-	.shrink = rpcauth_cache_shrinker,
+	.count_objects = rpcauth_cache_shrink_count,
+	.scan_objects = rpcauth_cache_shrink_scan,
 	.seeks = DEFAULT_SEEKS,
 };
 
@@ -897,8 +898,12 @@ int __init rpcauth_init_module(void)
 	err = rpc_init_authunix();
 	if (err < 0)
 		goto out1;
-	register_shrinker(&rpc_cred_shrinker);
+	err = register_shrinker(&rpc_cred_shrinker);
+	if (err < 0)
+		goto out2;
 	return 0;
+out2:
+	rpc_destroy_authunix();
 out1:
 	return err;
 }
@@ -908,3 +913,19 @@ void rpcauth_remove_module(void)
 	rpc_destroy_authunix();
 	unregister_shrinker(&rpc_cred_shrinker);
 }
+
+static int gid_cmp(const void *_a, const void *_b)
+{
+	kgid_t a = *(kgid_t *)_a;
+	kgid_t b = *(kgid_t *)_b;
+
+	return gid_gt(a, b) - gid_lt(a, b);
+}
+
+/* This is exported in the newer kernels */
+void sunrpc_groups_sort(struct group_info *group_info)
+{
+	sort(group_info->gid, group_info->ngroups, sizeof(*group_info->gid),
+	     gid_cmp, NULL);
+}
+EXPORT_SYMBOL(sunrpc_groups_sort);
