@@ -695,7 +695,7 @@ struct rpc_clnt *rpc_create(struct rpc_create_args *args)
 		switch (args->address->sa_family) {
 		case AF_LOCAL:
 			snprintf(servername, sizeof(servername), "%s",
-				 (const char *)(&sun->sun_path[0]));
+				 sun->sun_path);
 			break;
 		case AF_INET:
 			snprintf(servername, sizeof(servername), "%pI4",
