@@ -20,6 +20,7 @@
 #include <linux/spinlock.h>
 #include <linux/mutex.h>
 #include <linux/freezer.h>
+#include <linux/sched/mm.h>
 
 #include <linux/sunrpc/clnt.h>
 #include <linux/sunrpc/metrics.h>
@@ -992,10 +993,10 @@ void rpc_execute(struct rpc_task *task)
 
 static void rpc_async_schedule(struct work_struct *work)
 {
-	unsigned int pflags = memalloc_noio_save();
+	unsigned int pflags = memalloc_nofs_save();
 
 	__rpc_execute(container_of(work, struct rpc_task, u.tk_work));
-	memalloc_noio_restore(pflags);
+	memalloc_nofs_restore(pflags);
 }
 
 /**
@@ -1159,10 +1160,10 @@ static void rpc_free_task(struct rpc_task *task)
 
 static void rpc_async_release(struct work_struct *work)
 {
-	unsigned int pflags = memalloc_noio_save();
+	unsigned int pflags = memalloc_nofs_save();
 
 	rpc_free_task(container_of(work, struct rpc_task, u.tk_work));
-	memalloc_noio_restore(pflags);
+	memalloc_nofs_restore(pflags);
 }
 
 static void rpc_release_resources_task(struct rpc_task *task)
