@@ -17,6 +17,7 @@
 #include <linux/spinlock.h>
 #include <linux/sunrpc/clnt.h>
 #include <linux/sunrpc/xprt.h>
+#include <linux/sunrpc/kref.h>
 #include <linux/sunrpc/addr.h>
 #include <linux/sunrpc/xprtmultipath.h>
 
@@ -548,7 +549,7 @@ struct rpc_xprt_switch *xprt_iter_xchg_switch(struct rpc_xprt_iter *xpi,
 	struct rpc_xprt_switch __rcu *oldswitch;
 
 	/* Atomically swap out the old xpswitch */
-	oldswitch = xchg(&xpi->xpi_xpswitch, newswitch);
+	oldswitch = xchg(&xpi->xpi_xpswitch, RCU_INITIALIZER(newswitch));
 	if (newswitch != NULL)
 		xprt_iter_rewind(xpi);
 	return rcu_dereference_protected(oldswitch, true);
