@@ -49,13 +49,14 @@
  *  o buffer memory
  */
 
+#include "overflow.h"
+
 #include <linux/interrupt.h>
 #include <linux/slab.h>
 #include <linux/sunrpc/addr.h>
 #include <linux/sunrpc/svc_rdma.h>
 #include <linux/log2.h>
 #include <linux/version.h>
-#include <linux/module.h>
 
 #include <rdma/ib_cm.h>
 
